@@ -14,6 +14,7 @@
 #define LOG_ENABLE_DBG 0
 #include "log.h"
 #include "char32.h"
+#include "icon.h"
 #include "stride.h"
 #include "wayland.h"
 
@@ -21,7 +22,7 @@
 #define max(x, y) ((x) > (y) ? (x) : (y))
 
 struct render {
-    struct render_options options;
+    const struct config *conf;
     struct fcft_font *font;
     enum fcft_subpixel subpixel;
 
@@ -29,6 +30,14 @@ struct render {
     float dpi;
     bool size_font_by_dpi;
 
+    pixman_color_t pix_background_color;
+    pixman_color_t pix_border_color;
+    pixman_color_t pix_text_color;
+    pixman_color_t pix_match_color;
+    pixman_color_t pix_selection_color;
+    pixman_color_t pix_selection_text_color;
+    pixman_color_t pix_selection_match_color;
+
     unsigned x_margin;
     unsigned y_margin;
     unsigned inner_pad;
@@ -71,22 +80,22 @@ render_background(const struct render *render, struct buffer *buf)
 {
     bool use_pixman =
 #if defined(FUZZEL_ENABLE_CAIRO)
-        render->options.border_radius == 0
+        render->conf->border.radius == 0
 #else
         true
 #endif
         ;
 
     if (use_pixman) {
-        unsigned bw = render->options.border_size;
+        unsigned bw = render->conf->border.size;
 
-        pixman_color_t bg = rgba2pixman(render->options.background_color);
+        pixman_color_t bg = rgba2pixman(render->conf->colors.background);
         pixman_image_fill_rectangles(
             PIXMAN_OP_SRC, buf->pix, &bg,
             1, &(pixman_rectangle16_t){
                 bw, bw, buf->width - 2 * bw, buf->height - 2 * bw});
 
-        pixman_color_t border_color = rgba2pixman(render->options.border_color);
+        pixman_color_t border_color = rgba2pixman(render->conf->colors.border);
         pixman_image_fill_rectangles(
             PIXMAN_OP_SRC, buf->pix, &border_color,
             4, (pixman_rectangle16_t[]){
@@ -116,7 +125,7 @@ render_background(const struct render *render, struct buffer *buf)
         const double h = max(buf->height - 2 * b, 0.);
 
         const double from_degree = M_PI / 180;
-        const double radius = render->options.border_radius;
+        const double radius = render->conf->border.radius;
 
         /* Path describing an arc:ed rectangle */
         cairo_new_path(buf->cairo);
@@ -131,14 +140,14 @@ render_background(const struct render *render, struct buffer *buf)
         cairo_close_path(buf->cairo);
 
         /* Border */
-        const struct rgba *bc = &render->options.border_color;
+        const struct rgba *bc = &render->conf->colors.border;
         cairo_set_operator(buf->cairo, CAIRO_OPERATOR_SOURCE);
         cairo_set_line_width(buf->cairo, 2 * b);
         cairo_set_source_rgba(buf->cairo, bc->r, bc->g, bc->b, bc->a);
         cairo_stroke_preserve(buf->cairo);
 
         /* Background */
-        const struct rgba *bg = &render->options.background_color;
+        const struct rgba *bg = &render->conf->colors.background;
         cairo_set_source_rgba(buf->cairo, bg->r, bg->g, bg->b, bg->a);
         cairo_fill(buf->cairo);
 #else
@@ -174,26 +183,61 @@ render_prompt(const struct render *render, struct buffer *buf,
     struct fcft_font *font = render->font;
     assert(font != NULL);
 
+    const struct config *conf = render->conf;
+
     const char32_t *pprompt = prompt_prompt(prompt);
-    const size_t prompt_len = c32len(pprompt);
+    size_t prompt_len = c32len(pprompt);
 
     const char32_t *ptext = prompt_text(prompt);
     const size_t text_len = c32len(ptext);
 
     const enum fcft_subpixel subpixel =
-        (render->options.background_color.a == 1. &&
-         render->options.selection_color.a == 1.)
+        (render->conf->colors.background.a == 1. &&
+         render->conf->colors.selection.a == 1.)
         ? render->subpixel : FCFT_SUBPIXEL_NONE;
 
     int x = render->border_size + render->x_margin;
     int y = render->border_size + render->y_margin + font->ascent;
 
+    if (fcft_capabilities() & FCFT_CAPABILITY_TEXT_RUN_SHAPING) {
+        struct fcft_text_run *run = fcft_rasterize_text_run_utf32(
+            font, prompt_len, pprompt, subpixel);
+
+        if (run != NULL) {
+            for (size_t i = 0; i < run->count; i++) {
+                const struct fcft_glyph *glyph = run->glyphs[i];
+                render_glyph(buf->pix, glyph, x, y, &render->pix_text_color);
+                x += glyph->advance.x;
+            }
+            fcft_text_run_destroy(run);
+
+            /* Cursor, if right after the prompt. In all other cases,
+             * the cursor will be rendered by the loop below */
+            if (prompt_cursor(prompt) == 0) {
+                pixman_image_fill_rectangles(
+                    PIXMAN_OP_SRC, buf->pix, &render->pix_text_color,
+                    1, &(pixman_rectangle16_t){
+                        x, y - font->ascent,
+                        font->underline.thickness, font->ascent + font->descent});
+            }
+
+            /* Prevent loop below from rendering the prompt */;
+            prompt_len = 0;
+        }
+    }
+
     char32_t prev = 0;
 
     for (size_t i = 0; i < prompt_len + text_len; i++) {
-        char32_t wc = i < prompt_len ? pprompt[i] : ptext[i - prompt_len];
+        char32_t wc = i < prompt_len
+            ? pprompt[i]
+            : (conf->password != 0
+               ? conf->password
+               : ptext[i - prompt_len]);
+
         const struct fcft_glyph *glyph = fcft_rasterize_char_utf32(
             font, wc, subpixel);
+
         if (glyph == NULL) {
             prev = wc;
             continue;
@@ -203,15 +247,15 @@ render_prompt(const struct render *render, struct buffer *buf,
         fcft_kerning(font, prev, wc, &x_kern, NULL);
 
         x += x_kern;
-        render_glyph(buf->pix, glyph, x, y, &render->options.pix_text_color);
+        render_glyph(buf->pix, glyph, x, y, &render->pix_text_color);
         x += glyph->advance.x;
         if (i >= prompt_len)
-            x += pt_or_px_as_pixels(render, &render->options.letter_spacing);
+            x += pt_or_px_as_pixels(render, &render->conf->letter_spacing);
 
         /* Cursor */
         if (prompt_cursor(prompt) + prompt_len - 1 == i) {
             pixman_image_fill_rectangles(
-                PIXMAN_OP_SRC, buf->pix, &render->options.pix_text_color,
+                PIXMAN_OP_SRC, buf->pix, &render->pix_text_color,
                 1, &(pixman_rectangle16_t){
                     x, y - font->ascent,
                     font->underline.thickness, font->ascent + font->descent});
@@ -223,7 +267,8 @@ render_prompt(const struct render *render, struct buffer *buf,
 
 static void
 render_match_text(struct buffer *buf, double *_x, double _y, double max_x,
-                  const char32_t *text, ssize_t start, size_t length,
+                  const char32_t *text, size_t match_count,
+                  const struct match_substring matches[static match_count],
                   struct fcft_font *font, enum fcft_subpixel subpixel,
                   int letter_spacing,
                   pixman_color_t regular_color, pixman_color_t match_color,
@@ -273,7 +318,18 @@ render_match_text(struct buffer *buf, double *_x, double _y, double max_x,
     }
 
     for (size_t i = 0; i < count; i++) {
-        bool is_match = start >= 0 && clusters[i] >= start && clusters[i] < start + length;
+        bool is_match = false;
+        for (size_t j = 0; j < match_count; j++) {
+            const struct match_substring *match = &matches[j];
+            assert(match->start >= 0);
+
+            if (clusters[i] >= match->start &&
+                clusters[i] < match->start + match->len)
+            {
+                is_match = true;
+                break;
+            }
+        }
 
         if (x + (kern != NULL ? kern[i] : 0) + glyphs[i]->advance.x >= max_x) {
             const struct fcft_glyph *ellipses =
@@ -424,6 +480,11 @@ render_svg(struct icon *icon, int x, int y, int size, struct buffer *buf)
 {
     assert(icon->type == ICON_SVG);
 
+    if (icon->svg == NULL) {
+        if (!icon_from_svg(icon, icon->path))
+            return;
+    }
+
 #if defined(FUZZEL_ENABLE_SVG_LIBRSVG)
     render_svg_librsvg(icon, x, y, size, buf);
 #elif defined(FUZZEL_ENABLE_SVG_NANOSVG)
@@ -505,6 +566,11 @@ render_png(struct icon *icon, int x, int y, int size, struct buffer *buf)
 {
     assert(icon->type == ICON_PNG);
 
+    if (icon->png == NULL) {
+        if (!icon_from_png(icon, icon->path))
+            return;
+    }
+
 #if defined(FUZZEL_ENABLE_PNG_LIBPNG)
     render_png_libpng(icon, x, y, size, buf);
 #endif
@@ -524,8 +590,8 @@ render_match_list(const struct render *render, struct buffer *buf,
     const size_t match_count = matches_get_count(matches);
     const size_t selected = matches_get_match_index(matches);
     const enum fcft_subpixel subpixel =
-        (render->options.background_color.a == 1. &&
-         render->options.selection_color.a == 1.)
+        (render->conf->colors.background.a == 1. &&
+         render->conf->colors.selection.a == 1.)
         ? render->subpixel : FCFT_SUBPIXEL_NONE;
     const struct fcft_glyph *ellipses =
         fcft_rasterize_char_utf32(font, U'…', subpixel);
@@ -554,7 +620,8 @@ render_match_list(const struct render *render, struct buffer *buf,
              * there's "enough" free space, render a large
              * representation of the icon */
 
-            const double size = min(buf->height * 0.618, buf->width * 0.618);
+            const double ratio = render->conf->image_size_ratio;
+            const double size = min(buf->height * ratio, buf->width * ratio);
             const double img_x = (buf->width - size) / 2.;
             const double img_y_bottom = max(buf->height - first_row, 0.);
             const double img_y = max(img_y_bottom - size, 0.);
@@ -570,7 +637,7 @@ render_match_list(const struct render *render, struct buffer *buf,
                 render_svg(&match->application->icon, img_x, img_y, size, buf);
             }
 
-            pixman_color_t sc = rgba2pixman(render->options.selection_color);
+            pixman_color_t sc = rgba2pixman(render->conf->colors.selection);
             pixman_image_fill_rectangles(
                 PIXMAN_OP_SRC, buf->pix, &sc, 1,
                 &(pixman_rectangle16_t){
@@ -585,7 +652,7 @@ render_match_list(const struct render *render, struct buffer *buf,
         double max_x = buf->width - border_size - x_margin;
 
 #if 0 /* Render the icon+text bounding box */
-        pixman_color_t sc = rgba2pixman(render->options.match_color);
+        pixman_color_t sc = rgba2pixman(render->conf->match_color);
         pixman_image_fill_rectangles(
             PIXMAN_OP_SRC, buf->pix, &sc, 1,
             &(pixman_rectangle16_t){
@@ -617,20 +684,26 @@ render_match_list(const struct render *render, struct buffer *buf,
         const struct fcft_glyph *space = fcft_rasterize_char_utf32(
             font, U' ', subpixel);
 
-        cur_x += row_height +
-            (space != NULL ? space->advance.x : font->max_advance.x) +
-            pt_or_px_as_pixels(render, &render->options.letter_spacing);
+        cur_x +=
+            (render->conf->icons_enabled && matches_have_icons(matches)
+             ? (row_height +
+                (space != NULL ? space->advance.x : font->max_advance.x))
+             : 0) +
+            pt_or_px_as_pixels(render, &render->conf->letter_spacing);
 
         /* Application title */
         render_match_text(
             buf, &cur_x, y, max_x - (ellipses != NULL ? ellipses->width : 0),
-            match->application->title, match->start_title, c32len(prompt_text(prompt)),
+            match->application->title, match->pos_count, match->pos,
             font, subpixel,
-            pt_or_px_as_pixels(render, &render->options.letter_spacing),
+            pt_or_px_as_pixels(render, &render->conf->letter_spacing),
             (i == selected
-             ? render->options.pix_selection_text_color
-             : render->options.pix_text_color),
-            render->options.pix_match_color, &match->application->shaped);
+             ? render->pix_selection_text_color
+             : render->pix_text_color),
+            (i == selected
+             ? render->pix_selection_match_color
+             : render->pix_match_color),
+            &match->application->shaped);
 
         y += row_height;
     }
@@ -640,21 +713,21 @@ render_match_list(const struct render *render, struct buffer *buf,
 }
 
 struct render *
-render_init(const struct render_options *options, mtx_t *icon_lock)
+render_init(const struct config *conf, mtx_t *icon_lock)
 {
     struct render *render = calloc(1, sizeof(*render));
     *render = (struct render){
-        .options = *options,
+        .conf = conf,
         .icon_lock = icon_lock,
     };
 
-    /* TODO: the one providing the options should calculate these */
-    render->options.pix_background_color = rgba2pixman(render->options.background_color);
-    render->options.pix_border_color = rgba2pixman(render->options.border_color);
-    render->options.pix_text_color = rgba2pixman(render->options.text_color);
-    render->options.pix_match_color = rgba2pixman(render->options.match_color);
-    render->options.pix_selection_color = rgba2pixman(render->options.selection_color);
-    render->options.pix_selection_text_color = rgba2pixman(render->options.selection_text_color);
+    render->pix_background_color = rgba2pixman(render->conf->colors.background);
+    render->pix_border_color = rgba2pixman(render->conf->colors.border);
+    render->pix_text_color = rgba2pixman(render->conf->colors.text);
+    render->pix_match_color = rgba2pixman(render->conf->colors.match);
+    render->pix_selection_color = rgba2pixman(render->conf->colors.selection);
+    render->pix_selection_text_color = rgba2pixman(render->conf->colors.selection_text);
+    render->pix_selection_match_color = rgba2pixman(render->conf->colors.selection_match);
     return render;
 }
 
@@ -684,14 +757,14 @@ render_set_font(struct render *render, struct fcft_font *font,
     const struct fcft_glyph *W = fcft_rasterize_char_utf32(
         font, U'W', render->subpixel);
 
-    const unsigned x_margin = render->options.pad.x * scale;
-    const unsigned y_margin = render->options.pad.y * scale;
-    const unsigned inner_pad = render->options.pad.inner * scale;
+    const unsigned x_margin = render->conf->pad.x * scale;
+    const unsigned y_margin = render->conf->pad.y * scale;
+    const unsigned inner_pad = render->conf->pad.inner * scale;
 
-    const unsigned border_size = render->options.border_size * scale;
+    const unsigned border_size = render->conf->border.size * scale;
 
-    const unsigned row_height = render->options.line_height.px >= 0
-        ? pt_or_px_as_pixels(render, &render->options.line_height)
+    const unsigned row_height = render->conf->line_height.px >= 0
+        ? pt_or_px_as_pixels(render, &render->conf->line_height)
         : font->height;
 
     const unsigned icon_height = max(0, row_height - font->descent);
@@ -701,7 +774,7 @@ render_set_font(struct render *render, struct fcft_font *font,
         y_margin +
         row_height +                         /* The prompt */
         inner_pad +                          /* Padding between prompt and matches */
-        render->options.lines * row_height + /* Matches */
+        render->conf->lines * row_height + /* Matches */
         y_margin +
         border_size;                         /* Bottom border */
 
@@ -709,9 +782,9 @@ render_set_font(struct render *render, struct fcft_font *font,
         border_size +
         x_margin +
         (max((W->advance.x + pt_or_px_as_pixels(
-                  render, &render->options.letter_spacing)),
+                  render, &render->conf->letter_spacing)),
              0)
-         * render->options.chars) +
+         * render->conf->chars) +
         x_margin +
         border_size;
 
